# -*- coding: utf-8 -*-

import warnings
from typing import Union, Optional, Iterable

from matplotlib.ticker import MaxNLocator
from matplotlib import pyplot as plt
import matplotlib as mpl
from scipy import stats
import pandas as pd
import numpy as np

from lifelines.utils import coalesce, CensoringType, _group_event_table_by_intervals


__all__ = [
    "add_at_risk_counts",
    "plot_lifetimes",
    "plot_interval_censored_lifetimes",
    "qq_plot",
    "cdf_plot",
    "rmst_plot",
    "loglogs_plot",
]


def _iloc(x, i):
    """
    Returns the item at index i or items at indices i from x,
    where x is a numpy array or pd.Series.
    """
    try:
        return x.iloc[i]
    except AttributeError:
        return x[i]


def get_distribution_name_of_lifelines_model(model):
    return model._class_name.replace("Fitter", "").replace("AFT", "").lower()


def create_scipy_stats_model_from_lifelines_model(model):
    from lifelines.fitters import KnownModelParametricUnivariateFitter

    is_univariate_model = isinstance(model, KnownModelParametricUnivariateFitter)
    dist = get_distribution_name_of_lifelines_model(model)

    if not (is_univariate_model):
        raise TypeError(
            "Cannot use qq-plot with this model. See notes here: https://lifelines.readthedocs.io/en/latest/Examples.html?highlight=qq_plot#selecting-a-parametric-model-using-qq-plots"
        )

    if dist == "weibull":
        scipy_dist = "weibull_min"
        sparams = (model.rho_, 0, model.lambda_)

    elif dist == "lognormal":
        scipy_dist = "lognorm"
        sparams = (model.sigma_, 0, np.exp(model.mu_))

    elif dist == "loglogistic":
        scipy_dist = "fisk"
        sparams = (model.beta_, 0, model.alpha_)

    elif dist == "exponential":
        scipy_dist = "expon"
        sparams = (0, model.lambda_)

    else:
        raise NotImplementedError("Distribution not implemented in SciPy")

    return getattr(stats, scipy_dist)(*sparams)


def cdf_plot(model, timeline=None, ax=None, **plot_kwargs):
    """
    This plot compares the empirical CDF (derived by KaplanMeier) vs the model CDF.

    Parameters
    ------------
    model: lifelines univariate model
    timeline: iterable
    ax: matplotlib axis

    """
    from lifelines import KaplanMeierFitter

    if ax is None:
        ax = plt.gca()

    if timeline is None:
        timeline = model.timeline

    COL_EMP = "empirical CDF"

    if CensoringType.is_left_censoring(model):
        empirical_kmf = KaplanMeierFitter().fit_left_censoring(
            model.durations, model.event_observed, label=COL_EMP, timeline=timeline, weights=model.weights, entry=model.entry
        )
    elif CensoringType.is_right_censoring(model):
        empirical_kmf = KaplanMeierFitter().fit_right_censoring(
            model.durations, model.event_observed, label=COL_EMP, timeline=timeline, weights=model.weights, entry=model.entry
        )
    elif CensoringType.is_interval_censoring(model):
        empirical_kmf = KaplanMeierFitter().fit_interval_censoring(
            model.lower_bound, model.upper_bound, label=COL_EMP, timeline=timeline, weights=model.weights, entry=model.entry
        )

    empirical_kmf.plot_cumulative_density(ax=ax, **plot_kwargs)

    dist = get_distribution_name_of_lifelines_model(model)
    dist_object = create_scipy_stats_model_from_lifelines_model(model)
    ax.plot(timeline, dist_object.cdf(timeline), label="fitted %s" % dist, **plot_kwargs)
    ax.legend()
    return ax


def rmst_plot(model, model2=None, t=np.inf, ax=None, text_position=None, **plot_kwargs):
    """
    This functions plots the survival function of the model plus it's area-under-the-curve (AUC) up
    until the point ``t``. The AUC is known as the restricted mean survival time (RMST).

    To compare the difference between two models' survival curves, you can supply an
    additional model in ``model2``.

    Parameters
    -----------
    model: lifelines.UnivariateFitter
    model2: lifelines.UnivariateFitter, optional
        used to compute the delta RMST of two models
    t: float
        the upper bound of the expectation
    ax: axis
    text_position: tuple
        move the text position of the RMST.


    Examples
    ---------
    .. code:: python

        from lifelines.utils import restricted_mean_survival_time
        from lifelines.datasets import load_waltons
        from lifelines.plotting import rmst_plot

        df = load_waltons()
        ix = df['group'] == 'miR-137'
        T, E = df['T'], df['E']
        time_limit = 50

        kmf_exp = KaplanMeierFitter().fit(T[ix], E[ix], label='exp')
        kmf_con = KaplanMeierFitter().fit(T[~ix], E[~ix], label='control')

        ax = plt.subplot(311)
        rmst_plot(kmf_exp, t=time_limit, ax=ax)

        ax = plt.subplot(312)
        rmst_plot(kmf_con, t=time_limit, ax=ax)

        ax = plt.subplot(313)
        rmst_plot(kmf_exp, model2=kmf_con, t=time_limit, ax=ax)



    """
    from lifelines.utils import restricted_mean_survival_time

    if ax is None:
        ax = plt.gca()

    rmst = restricted_mean_survival_time(model, t=t)
    c = ax._get_lines.get_next_color()
    model.plot_survival_function(ax=ax, color=c, ci_show=False, **plot_kwargs)

    if text_position is None:
        text_position = (np.percentile(model.timeline, 10), 0.15)

    if model2 is not None:
        c2 = ax._get_lines.get_next_color()
        rmst2 = restricted_mean_survival_time(model2, t=t)
        model2.plot_survival_function(ax=ax, color=c2, ci_show=False, **plot_kwargs)
        timeline = np.unique(model.timeline.tolist() + model2.timeline.tolist() + [t])
        predict1 = model.predict(timeline).loc[:t]
        predict2 = model2.predict(timeline).loc[:t]
        # positive
        ax.fill_between(
            timeline[timeline <= t],
            predict1,
            predict2,
            where=predict1 > predict2,
            step="post",
            facecolor="w",
            hatch="|",
            edgecolor="grey",
        )

        # negative
        ax.fill_between(
            timeline[timeline <= t],
            predict1,
            predict2,
            where=predict1 < predict2,
            step="post",
            hatch="-",
            facecolor="w",
            edgecolor="grey",
        )

        ax.text(
            text_position[0], text_position[1], "RMST(%s) -\n   RMST(%s)=%.3f" % (model._label, model2._label, rmst - rmst2)
        )  # dynamically pick this.
    else:
        rmst = restricted_mean_survival_time(model, t=t)
        sf_exp_at_limit = model.predict(np.append(model.timeline, t)).sort_index().loc[:t]
        ax.fill_between(sf_exp_at_limit.index, sf_exp_at_limit.values, step="post", color=c, alpha=0.25)
        ax.text(text_position[0], text_position[1], "RMST=%.3f" % rmst)  # dynamically pick this.

    ax.axvline(t, ls="--", color="k")
    ax.set_ylim(0, 1)
    return ax


def qq_plot(model, ax=None, **plot_kwargs):
    """
    Produces a quantile-quantile plot of the empirical CDF against
    the fitted parametric CDF. Large deviances away from the line y=x
    can invalidate a model (though we expect some natural deviance in the tails).

    Parameters
    -----------
    model: obj
        A fitted lifelines univariate parametric model, like ``WeibullFitter``
    plot_kwargs:
        kwargs for the plot.

    Returns
    --------
    ax:
        The axes which was used.

    Examples
    ---------
    .. code:: python

        from lifelines import *
        from lifelines.plotting import qq_plot
        from lifelines.datasets import load_rossi
        df = load_rossi()
        wf = WeibullFitter().fit(df['week'], df['arrest'])
        qq_plot(wf)

    Notes
    ------
    The interval censoring case uses the mean between the upper and lower bounds.

    """
    from lifelines.utils import qth_survival_times
    from lifelines import KaplanMeierFitter

    if ax is None:
        ax = plt.gca()

    dist = get_distribution_name_of_lifelines_model(model)
    dist_object = create_scipy_stats_model_from_lifelines_model(model)

    COL_EMP = "empirical quantiles"
    COL_THEO = "fitted %s quantiles" % dist

    if CensoringType.is_left_censoring(model):
        kmf = KaplanMeierFitter().fit_left_censoring(
            model.durations, model.event_observed, label=COL_EMP, weights=model.weights, entry=model.entry
        )
        sf, cdf = kmf.survival_function_[COL_EMP], kmf.cumulative_density_[COL_EMP]
    elif CensoringType.is_right_censoring(model):
        kmf = KaplanMeierFitter().fit_right_censoring(
            model.durations, model.event_observed, label=COL_EMP, weights=model.weights, entry=model.entry
        )
        sf, cdf = kmf.survival_function_[COL_EMP], kmf.cumulative_density_[COL_EMP]

    elif CensoringType.is_interval_censoring(model):
        kmf = KaplanMeierFitter().fit_interval_censoring(
            model.lower_bound, model.upper_bound, label=COL_EMP, weights=model.weights, entry=model.entry
        )
        sf, cdf = kmf.survival_function_.mean(1), kmf.cumulative_density_[COL_EMP + "_lower"]

    q = np.unique(cdf.values)

    quantiles = qth_survival_times(1 - q, sf)
    quantiles[COL_THEO] = dist_object.ppf(q)
    quantiles = quantiles.replace([-np.inf, 0, np.inf], np.nan).dropna()

    max_, min_ = quantiles[COL_EMP].max(), quantiles[COL_EMP].min()

    quantiles.plot.scatter(COL_THEO, COL_EMP, c="none", edgecolor="k", lw=0.5, ax=ax)
    ax.plot([min_, max_], [min_, max_], c="k", ls=":", lw=1.0)
    ax.set_ylim(min_, max_)
    ax.set_xlim(min_, max_)

    return ax


def is_latex_enabled():
    """
    Returns True if LaTeX is enabled in matplotlib's rcParams,
    False otherwise
    """
    import matplotlib as mpl

    return mpl.rcParams["text.usetex"]


def remove_spines(ax, sides):
    """
    Remove spines of axis.

    Parameters:
      ax: axes to operate on
      sides: list of sides: top, left, bottom, right

    Examples:
    removespines(ax, ['top'])
    removespines(ax, ['top', 'bottom', 'right', 'left'])
    """
    for side in sides:
        ax.spines[side].set_visible(False)
    return ax


def move_spines(ax, sides, dists):
    """
    Move the entire spine relative to the figure.

    Parameters:
      ax: axes to operate on
      sides: list of sides to move. Sides: top, left, bottom, right
      dists: list of float distances to move. Should match sides in length.

    Example:
    move_spines(ax, sides=['left', 'bottom'], dists=[-0.02, 0.1])
    """
    for side, dist in zip(sides, dists):
        ax.spines[side].set_position(("axes", dist))
    return ax


def remove_ticks(ax, x=False, y=False):
    """
    Remove ticks from axis.

    Parameters:
      ax: axes to work on
      x: if True, remove xticks. Default False.
      y: if True, remove yticks. Default False.

    Examples:
    removeticks(ax, x=True)
    removeticks(ax, x=True, y=True)
    """
    if x:
        ax.xaxis.set_ticks_position("none")
    if y:
        ax.yaxis.set_ticks_position("none")
    return ax


def add_at_risk_counts(*fitters, labels: Optional[Union[Iterable, bool]] = None, rows_to_show=None, ypos=-0.6, ax=None, **kwargs):
    """
    Add counts showing how many individuals were at risk, censored, and observed, at each time point in
    survival/hazard plots.

    Tip: you may want to call ``plt.tight_layout()`` afterwards.

    Parameters
    ----------
    fitters:
      One or several fitters, for example KaplanMeierFitter, WeibullFitter,
      NelsonAalenFitter, etc...
    ax:
        a matplotlib axes
    labels:
        provide labels for the fitters, default is to use the provided fitter label. Set to
        False for no labels.
    rows_to_show: list
        list of a subset of {'At risk', 'Censored', 'Events'}. Default shows all columns.
    ypos:
        increase to move the table down further.

    Returns
    --------
      ax:
        The axes which was used.

    Examples
    --------
    .. code:: python

        # First train some fitters and plot them
        fig = plt.figure()
        ax = plt.subplot(111)

        f1 = KaplanMeierFitter()
        f1.fit(data)
        f1.plot(ax=ax)

        f2 = KaplanMeierFitter()
        f2.fit(data)
        f2.plot(ax=ax)

        # There are equivalent
        add_at_risk_counts(f1, f2)
        add_at_risk_counts(f1, f2, ax=ax, fig=fig)

        # This overrides the labels
        add_at_risk_counts(f1, f2, labels=['fitter one', 'fitter two'])

        # This hides the labels
        add_at_risk_counts(f1, f2, labels=False)

    References
    -----------
     Morris TP, Jarvis CI, Cragg W, et al. Proposals on Kaplan–Meier plots in medical research and a survey of stakeholder views: KMunicate. BMJ Open 2019;9:e030215. doi:10.1136/bmjopen-2019-030215

    """

    if ax is None:
        ax = plt.gca()

    fig = kwargs.pop("fig", None)
    if fig is None:
        fig = plt.gcf()

    if labels is None:
        labels = [f._label for f in fitters]
    elif labels is False:
        labels = [None] * len(fitters)

    if rows_to_show is None:
        rows_to_show = ["At risk", "Censored", "Events"]
    n_rows = len(rows_to_show)

    # Create another axes where we can put size ticks
    ax2 = plt.twiny(ax=ax)
    # Move the ticks below existing axes
    # Appropriate length scaled for 6 inches. Adjust for figure size.
    ax_height = (ax.get_position().y1 - ax.get_position().y0) * fig.get_figheight()  # axis height
    ax2_ypos = ypos / ax_height

    move_spines(ax2, ["bottom"], [ax2_ypos])
    # Hide all fluff
    remove_spines(ax2, ["top", "right", "bottom", "left"])
    # Set ticks and labels on bottom
    ax2.xaxis.tick_bottom()
    # Set limit
    min_time, max_time = ax.get_xlim()
    ax2.set_xlim(min_time, max_time)
    # Set ticks to kwarg or visible ticks
    xticks = kwargs.pop("xticks", None)
    if xticks is None:
        xticks = [xtick for xtick in ax.get_xticks() if min_time <= xtick <= max_time]
    ax2.set_xticks(xticks)
    # Remove ticks, need to do this AFTER moving the ticks
    remove_ticks(ax2, x=True, y=True)

    ticklabels = []

    for tick in ax2.get_xticks():
        lbl = ""

        # Get counts at tick
        counts = []
        for f in fitters:
            # this is a messy:
            # a) to align with R (and intuition), we do a subtraction off the at_risk column
            # b) we group by the tick intervals
            # c) we want to start at 0, so we give it it's own interval

            event_table_slice = (
                f.event_table.assign(at_risk=lambda x: x.at_risk - x.removed)
                .loc[:tick, ["at_risk", "censored", "observed"]]
                .agg({"at_risk": "min", "censored": "sum", "observed": "sum"})
                .rename({"at_risk": "At risk", "censored": "Censored", "observed": "Events"})
            )
            counts.extend([int(c) for c in event_table_slice.loc[rows_to_show]])

        if tick == ax2.get_xticks()[0]:
            max_length = len(str(max(counts)))
            for i, c in enumerate(counts):
                if i % n_rows == 0:
                    if is_latex_enabled():
                        lbl += ("\n" if i > 0 else "") + r"\textbf{%s}" % labels[int(i / n_rows)] + "\n"
                    else:
                        lbl += ("\n" if i > 0 else "") + r"%s" % labels[int(i / n_rows)] + "\n"

                l = rows_to_show[i % n_rows]
                l = {"At risk": "  At risk", "Censored": "Censored  ", "Events": " Events  "}.get(l)
                s = "{}   ".format(l.rjust(10, " ")) + "{{:>{}d}}\n".format(max_length)
                lbl += s.format(c)

        else:
            # Create tick label
            lbl += ""
            for i, c in enumerate(counts):
                if i % n_rows == 0 and i > 0:
                    lbl += "\n\n"
                s = "\n{}"
                lbl += s.format(c)

        ticklabels.append(lbl)

    # Align labels to the right so numbers can be compared easily
    ax2.set_xticklabels(ticklabels, ha="right", **kwargs)

    return ax


def plot_interval_censored_lifetimes(
    lower_bound,
    upper_bound,
    entry=None,
    left_truncated=False,
    sort_by_lower_bound=True,
    event_observed_color="#A60628",
    event_right_censored_color="#348ABD",
    ax=None,
    **kwargs
):
    """
    Returns a lifetime plot for interval censored data.

    Parameters
    -----------
    lower_bound: (n,) numpy array or pd.Series
      the start of the period the subject experienced the event in.
    upper_bound: (n,) numpy array or pd.Series
      the end of the period the subject experienced the event in. If the value is equal to the corresponding value in lower_bound, then
      the individual's event was observed (not censored).
    entry: (n,) numpy array or pd.Series
      offsetting the births away from t=0. This could be from left-truncation, or delayed entry into study.
    left_truncated: boolean
      if entry is provided, and the data is left-truncated, this will display additional information in the plot to reflect this.
    sort_by_lower_bound: boolean
      sort by the lower_bound vector
    event_observed_color: str
      default: "#A60628"
    event_right_censored_color: str
      default: "#348ABD"
      applies to any individual with an upper bound of infinity.

    Returns
    -------
    ax:

    Examples
    ---------
    .. code:: python

        import pandas as pd
        import numpy as np
        from lifelines.plotting import plot_interval_censored_lifetimes
        df = pd.DataFrame({'lb':[20,15,30, 10, 20, 30], 'ub':[25, 15, np.infty, 20, 20, np.infty]})
        ax = plot_interval_censored_lifetimes(lower_bound=df['lb'], upper_bound=df['ub'])
    """

    if ax is None:
        ax = plt.gca()

    # If lower_bounds is pd.Series with non-default index, then use index values as y-axis labels.
    label_plot_bars = type(lower_bound) is pd.Series and type(lower_bound.index) is not pd.RangeIndex

    N = lower_bound.shape[0]
    if N > 25:
        warnings.warn("For less visual clutter, you may want to subsample to less than 25 individuals.")

    assert upper_bound.shape[0] == N

    if sort_by_lower_bound:
        ix = np.argsort(lower_bound, 0)
        upper_bound = _iloc(upper_bound, ix)
        lower_bound = _iloc(lower_bound, ix)
        if entry:
            entry = _iloc(lower_bound, ix)

    if entry is None:
        entry = np.zeros(N)

    for i in range(N):
        if np.isposinf(_iloc(upper_bound, i)):
            c = event_right_censored_color
            ax.hlines(i, _iloc(entry, i), _iloc(lower_bound, i), color=c, lw=1.5)
        else:
            c = event_observed_color
            ax.hlines(i, _iloc(entry, i), _iloc(upper_bound, i), color=c, lw=1.5)
            if _iloc(lower_bound, i) == _iloc(upper_bound, i):
                ax.scatter(_iloc(lower_bound, i), i, color=c, marker="o", s=13)
            else:
                ax.scatter(_iloc(lower_bound, i), i, color=c, marker=">", s=13)
                ax.scatter(_iloc(upper_bound, i), i, color=c, marker="<", s=13)

        if left_truncated:
            ax.hlines(i, 0, _iloc(entry, i), color=c, lw=1.0, linestyle="--")

    if label_plot_bars:
        ax.set_yticks(range(0, N))
        ax.set_yticklabels(lower_bound.index)
    else:
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    ax.set_xlim(0)
    ax.set_ylim(-0.5, N)
    return ax


def plot_lifetimes(
    durations,
    event_observed=None,
    entry=None,
    left_truncated=False,
    sort_by_duration=True,
    event_observed_color="#A60628",
    event_censored_color="#348ABD",
    ax=None,
    **kwargs
):
    """
    Returns a lifetime plot, see examples: https://lifelines.readthedocs.io/en/latest/Survival%20Analysis%20intro.html#Censoring

    Parameters
    -----------
    durations: (n,) numpy array or pd.Series
      duration subject was observed for.
    event_observed: (n,) numpy array or pd.Series
      array of booleans: True if event observed, else False.
    entry: (n,) numpy array or pd.Series
      offsetting the births away from t=0. This could be from left-truncation, or delayed entry into study.
    left_truncated: boolean
      if entry is provided, and the data is left-truncated, this will display additional information in the plot to reflect this.
    sort_by_duration: boolean
      sort by the duration vector
    event_observed_color: str
      default: "#A60628"
    event_censored_color: str
      default: "#348ABD"

    Returns
    -------
    ax:

    Examples
    ---------
    .. code:: python

        from lifelines.datasets import load_waltons
        from lifelines.plotting import plot_lifetimes
        T, E = load_waltons()["T"], load_waltons()["E"]
        ax = plot_lifetimes(T.loc[:50], event_observed=E.loc[:50])

    """

    if ax is None:
        ax = plt.gca()

    # If durations is pd.Series with non-default index, then use index values as y-axis labels.
    label_plot_bars = type(durations) is pd.Series and type(durations.index) is not pd.RangeIndex

    N = durations.shape[0]
    if N > 25:
        warnings.warn("For less visual clutter, you may want to subsample to less than 25 individuals.")

    if event_observed is None:
        event_observed = np.ones(N, dtype=bool)

    if entry is None:
        entry = np.zeros(N)

    assert durations.shape[0] == N
    assert event_observed.shape[0] == N

    if sort_by_duration:
        # order by length of lifetimes;
        ix = np.argsort(entry + durations, 0)
        durations = _iloc(durations, ix)
        event_observed = _iloc(event_observed, ix)
        entry = _iloc(entry, ix)

    for i in range(N):
        c = event_observed_color if _iloc(event_observed, i) else event_censored_color
        ax.hlines(i, _iloc(entry, i), _iloc(entry, i) + _iloc(durations, i), color=c, lw=1.5)
        if left_truncated:
            ax.hlines(i, 0, _iloc(entry, i), color=c, lw=1.0, linestyle="--")
        m = "" if not _iloc(event_observed, i) else "o"
        ax.scatter(_iloc(entry, i) + _iloc(durations, i), i, color=c, marker=m, s=13)

    if label_plot_bars:
        ax.set_yticks(range(0, N))
        ax.set_yticklabels(durations.index)
    else:
        ax.yaxis.set_major_locator(MaxNLocator(integer=True))

    ax.set_xlim(0)
    ax.set_ylim(-0.5, N)
    return ax


def set_kwargs_color(kwargs):
    kwargs["color"] = coalesce(kwargs.get("c"), kwargs.get("color"), kwargs["ax"]._get_lines.get_next_color())


def set_kwargs_drawstyle(kwargs, default="steps-post"):
    kwargs["drawstyle"] = kwargs.get("drawstyle", default)


def set_kwargs_label(kwargs, cls):
    kwargs["label"] = kwargs.get("label", cls._label)


def create_dataframe_slicer(iloc, loc, timeline):
    if (loc is not None) and (iloc is not None):
        raise ValueError("Cannot set both loc and iloc in call to .plot().")

    user_did_not_specify_certain_indexes = (iloc is None) and (loc is None)
    user_submitted_slice = slice(timeline.min(), timeline.max()) if user_did_not_specify_certain_indexes else coalesce(loc, iloc)

    get_method = "iloc" if iloc is not None else "loc"
    return lambda df: getattr(df, get_method)[user_submitted_slice]


def loglogs_plot(cls, loc=None, iloc=None, show_censors=False, censor_styles=None, ax=None, **kwargs):
    """
    Specifies a plot of the log(-log(SV)) versus log(time) where SV is the estimated survival function.
    """

    def loglog(s):
        return np.log(-np.log(s))

    if (loc is not None) and (iloc is not None):
        raise ValueError("Cannot set both loc and iloc in call to .plot().")

    if censor_styles is None:
        censor_styles = {}

    if ax is None:
        ax = plt.gca()

    kwargs["ax"] = ax
    set_kwargs_color(kwargs)
    set_kwargs_drawstyle(kwargs)

    dataframe_slicer = create_dataframe_slicer(iloc, loc, cls.timeline)

    # plot censors
    colour = kwargs["color"]

    if show_censors and cls.event_table["censored"].sum() > 0:
        cs = {"marker": "|", "ms": 12, "mew": 1}
        cs.update(censor_styles)
        times = dataframe_slicer(cls.event_table.loc[(cls.event_table["censored"] > 0)]).index.values.astype(float)
        v = cls.predict(times)
        ax.plot(np.log(times), loglog(v), linestyle="None", color=colour, **cs)

    # plot estimate
    sliced_estimates = dataframe_slicer(loglog(cls.survival_function_))
    sliced_estimates["log(timeline)"] = np.log(sliced_estimates.index)
    sliced_estimates.plot(x="log(timeline)", **kwargs)
    ax.set_xlabel("log(timeline)")
    ax.set_ylabel("log(-log(survival_function_))")

    return ax


def _plot_estimate(
    cls,
    estimate=None,  # string like "survival_function_", "cumulative_density_", "hazard_", "cumulative_hazard_"
    loc=None,
    iloc=None,
    show_censors=False,
    censor_styles=None,
    ci_legend=False,
    ci_force_lines=False,
    ci_only_lines=False,
    ci_no_lines=False,
    ci_alpha=0.25,
    ci_show=True,
    at_risk_counts=False,
    logx: bool = False,
    ax=None,
    **kwargs
):

    """
    Plots a pretty figure of estimates

    Matplotlib plot arguments can be passed in inside the kwargs, plus

    Parameters
    -----------
    show_censors: bool
        place markers at censorship events. Default: False
    censor_styles: bool
        If show_censors, this dictionary will be passed into the plot call.
    ci_show: bool
        show confidence intervals. Default: True
    ci_alpha: bool
        the transparency level of the confidence interval. Default: 0.3
    ci_force_lines: bool
        make the confidence intervals to be line plots (versus default shaded areas + lines). Default: False
        Deprecated: use ``ci_only_lines`` instead.
    ci_only_lines: bool
        make the confidence intervals to be line plots (versus default shaded areas + lines). Default: False.
    ci_no_lines: bool
        Only show the shaded area, with no boarding lines. Default: False
    ci_legend: bool
        if ci_force_lines is True, this is a boolean flag to add the lines' labels to the legend. Default: False
    at_risk_counts: bool
        show group sizes at time points. See function ``add_at_risk_counts`` for details. Default: False
    loc: slice
        specify a time-based subsection of the curves to plot, ex:

        >>> model.plot(loc=slice(0.,10.))

        will plot the time values between t=0. and t=10.
    iloc: slice
        specify a location-based subsection of the curves to plot, ex:

        >>> model.plot(iloc=slice(0,10))

        will plot the first 10 time points.
    logx: bool
        Use log scaling on x axis


    Returns
    -------
    ax:
        a pyplot axis object
    """
    if ci_force_lines:
        warnings.warn(
            "ci_force_lines is deprecated. Use ci_only_lines instead (no functional difference, only a name change).",
            DeprecationWarning,
        )
        ci_only_lines = ci_force_lines

    plot_estimate_config = PlotEstimateConfig(cls, estimate, loc, iloc, show_censors, censor_styles, logx, ax, **kwargs)

    dataframe_slicer = create_dataframe_slicer(iloc, loc, cls.timeline)

    if show_censors and cls.event_table["censored"].sum() > 0:
        cs = {"marker": "+", "ms": 12, "mew": 1}
        cs.update(plot_estimate_config.censor_styles)
        censored_times = dataframe_slicer(cls.event_table.loc[(cls.event_table["censored"] > 0)]).index.values.astype(float)
        v = plot_estimate_config.predict_at_times(censored_times).values
        plot_estimate_config.ax.plot(censored_times, v, linestyle="None", color=plot_estimate_config.colour, **cs)

    dataframe_slicer(plot_estimate_config.estimate_).rename(columns=lambda _: plot_estimate_config.kwargs.pop("label")).plot(
        logx=plot_estimate_config.logx, **plot_estimate_config.kwargs
    )

    # plot confidence intervals
    if ci_show:
        if ci_only_lines:
            # see https://github.com/CamDavidsonPilon/lifelines/issues/928
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                (
                    dataframe_slicer(plot_estimate_config.confidence_interval_)
                    .rename(columns=lambda s: ("" if ci_legend else "_") + s)
                    .plot(
                        linestyle="-",
                        linewidth=1,
                        color=[plot_estimate_config.colour],
                        drawstyle=plot_estimate_config.kwargs["drawstyle"],
                        alpha=0.6,
                        logx=plot_estimate_config.logx,
                        ax=plot_estimate_config.ax,
                    )
                )
        else:
            x = dataframe_slicer(plot_estimate_config.confidence_interval_).index.values.astype(float)
            lower = dataframe_slicer(plot_estimate_config.confidence_interval_.iloc[:, [0]]).values[:, 0]
            upper = dataframe_slicer(plot_estimate_config.confidence_interval_.iloc[:, [1]]).values[:, 0]

            if plot_estimate_config.kwargs["drawstyle"] == "default":
                step = None
            elif plot_estimate_config.kwargs["drawstyle"].startswith("step"):
                step = plot_estimate_config.kwargs["drawstyle"].replace("steps-", "")

            plot_estimate_config.ax.fill_between(
                x,
                lower,
                upper,
                alpha=ci_alpha,
                color=plot_estimate_config.colour,
                linewidth=0.0 if ci_no_lines else 1.0,
                step=step,
            )

    if at_risk_counts:
        add_at_risk_counts(cls, ax=plot_estimate_config.ax)
        plt.tight_layout()

    return plot_estimate_config.ax


class PlotEstimateConfig:
    def __init__(self, cls, estimate: Union[str, pd.DataFrame], loc, iloc, show_censors, censor_styles, logx, ax, **kwargs):

        self.censor_styles = coalesce(censor_styles, {})

        if ax is None:
            ax = plt.gca()

        kwargs["ax"] = ax
        set_kwargs_color(kwargs)
        set_kwargs_drawstyle(kwargs)
        set_kwargs_label(kwargs, cls)

        self.loc = loc
        self.iloc = iloc
        self.show_censors = show_censors
        # plot censors
        self.ax = ax
        self.colour = kwargs["color"]
        self.logx = logx
        self.kwargs = kwargs

        if isinstance(estimate, str):
            self.estimate_ = getattr(cls, estimate)
            self.confidence_interval_ = getattr(cls, "confidence_interval_" + estimate)
            self.predict_at_times = getattr(cls, estimate + "at_times")
        else:
            self.estimate_ = estimate
            self.confidence_interval_ = kwargs.pop("confidence_intervals")



# -*- coding: utf-8 -*-
import warnings
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt

from lifelines.utils import CensoringType
from lifelines.fitters import RegressionFitter
from lifelines import CRCSplineFitter


def survival_probability_calibration(model: RegressionFitter, df: pd.DataFrame, t0: float, ax=None):
    r"""
    Smoothed calibration curves for time-to-event models. This is analogous to
    calibration curves for classification models, extended to handle survival probabilities
    and censoring. Produces a matplotlib figure and some metrics.

    We want to calibrate our model's prediction of :math:`P(T < \text{t0})` against the observed frequencies.

    Parameters
    -------------

    model:
        a fitted lifelines regression model to be evaluated
    df: DataFrame
        a DataFrame - if equal to the training data, then this is an in-sample calibration. Could also be an out-of-sample
        dataset.
    t0: float
        the time to evaluate the probability of event occurring prior at.

    Returns
    ----------
    ax:
        mpl axes
    ICI:
        mean absolute difference between predicted and observed
    E50:
        median absolute difference between predicted and observed

    https://onlinelibrary.wiley.com/doi/full/10.1002/sim.8570

    """

    def ccl(p):
        return np.log(-np.log(1 - p))

    if ax is None:
        ax = plt.gca()

    T = model.duration_col
    E = model.event_col

    predictions_at_t0 = np.clip(1 - model.predict_survival_function(df, times=[t0]).T.squeeze(), 1e-10, 1 - 1e-10)

    # create new dataset with the predictions
    prediction_df = pd.DataFrame({"ccl_at_%d" % t0: ccl(predictions_at_t0), T: df[T], E: df[E]})

    # fit new dataset to flexible spline model
    # this new model connects prediction probabilities and actual survival. It should be very flexible, almost to the point of overfitting. It's goal is just to smooth out the data!
    knots = 3
    regressors = {"beta_": ["ccl_at_%d" % t0], "gamma0_": "1", "gamma1_": "1", "gamma2_": "1"}

    # this model is from examples/royson_crowther_clements_splines.py
    crc = CRCSplineFitter(knots, penalizer=0.000001)
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore")
        if CensoringType.is_right_censoring(model):
            crc.fit_right_censoring(prediction_df, T, E, regressors=regressors)
        elif CensoringType.is_left_censoring(model):
            crc.fit_left_censoring(prediction_df, T, E, regressors=regressors)
        elif CensoringType.is_interval_censoring(model):
            crc.fit_interval_censoring(prediction_df, T, E, regressors=regressors)

    # predict new model at values 0 to 1, but remember to ccl it!
    x = np.linspace(np.clip(predictions_at_t0.min() - 0.01, 0, 1), np.clip(predictions_at_t0.max() + 0.01, 0, 1), 100)
    y = 1 - crc.predict_survival_function(pd.DataFrame({"ccl_at_%d" % t0: ccl(x)}), times=[t0]).T.squeeze()

    # plot our results
    ax.set_title("Smoothed calibration curve of \npredicted vs observed probabilities of t ≤ %d mortality" % t0)

    color = "tab:red"
    ax.plot(x, y, label="smoothed calibration curve", color=color)
    ax.set_xlabel("Predicted probability of \nt ≤ %d mortality" % t0)
    ax.set_ylabel("Observed probability of \nt ≤ %d mortality" % t0, color=color)
    ax.tick_params(axis="y", labelcolor=color)

    # plot x=y line
    ax.plot(x, x, c="k", ls="--")
    ax.legend()

    # plot histogram of our original predictions
    color = "tab:blue"
    twin_ax = ax.twinx()
    twin_ax.set_ylabel("Count of \npredicted probabilities", color=color)  # we already handled the x-label with ax1
    twin_ax.tick_params(axis="y", labelcolor=color)
    twin_ax.hist(predictions_at_t0, alpha=0.3, bins="sqrt", color=color)

    plt.tight_layout()

    deltas = ((1 - crc.predict_survival_function(prediction_df, times=[t0])).T.squeeze() - predictions_at_t0).abs()
    ICI = deltas.mean()
    E50 = np.percentile(deltas, 50)
    print("ICI = ", ICI)
    print("E50 = ", E50)

    return ax, ICI, E50



# -*- coding: utf-8 -*-


class StatError(Exception):
    pass


class ConvergenceError(ValueError):
    # inherits from ValueError for backwards compatibility reasons
    def __init__(self, msg, original_exception=""):
        super(ConvergenceError, self).__init__(msg + "%s" % original_exception)
        self.original_exception = original_exception


class ConvergenceWarning(RuntimeWarning):
    pass


class StatisticalWarning(RuntimeWarning):
    pass


class ApproximationWarning(RuntimeWarning):
    pass



# -*- coding: utf-8 -*-
from __future__ import print_function

import pytest
import numpy as np

from lifelines.utils.btree import _BTree as BTree


def test_btree():
    t = BTree(np.arange(10))
    for i in range(10):
        assert t.rank(i) == (0, 0)

    assert len(t) == 0
    t.insert(5)
    t.insert(6)
    t.insert(6)
    t.insert(0)
    t.insert(9)
    assert len(t) == 5

    assert t.rank(0) == (0, 1)
    assert t.rank(0.5) == (1, 0)
    assert t.rank(4.5) == (1, 0)
    assert t.rank(5) == (1, 1)
    assert t.rank(5.5) == (2, 0)
    assert t.rank(6) == (2, 2)
    assert t.rank(6.5) == (4, 0)
    assert t.rank(8.5) == (4, 0)
    assert t.rank(9) == (4, 1)
    assert t.rank(9.5) == (5, 0)

    for i in range(1, 32):
        BTree(np.arange(i))

    with pytest.raises(ValueError):
        # This has to go last since it screws up the counts
        t.insert(5.5)



# -*- coding: utf-8 -*-
import os
import pytest
import pandas as pd
import numpy as np
import scipy

from lifelines import (
    NelsonAalenFitter,
    KaplanMeierFitter,
    CoxPHFitter,
    CoxTimeVaryingFitter,
    AalenAdditiveFitter,
    WeibullFitter,
    LogNormalFitter,
    LogLogisticFitter,
    WeibullAFTFitter,
    ExponentialFitter,
    AalenJohansenFitter,
    BreslowFlemingHarringtonFitter,
)

from lifelines.tests.test_estimation import known_parametric_univariate_fitters

from lifelines.generate_datasets import generate_random_lifetimes, generate_hazard_rates
from lifelines.plotting import plot_lifetimes, cdf_plot, qq_plot, rmst_plot, add_at_risk_counts
from lifelines.datasets import (
    load_waltons,
    load_regression_dataset,
    load_lcd,
    load_panel_test,
    load_stanford_heart_transplants,
    load_rossi,
    load_multicenter_aids_cohort_study,
    load_nh4,
    load_diabetes,
)
from lifelines.generate_datasets import cumulative_integral
from lifelines.calibration import survival_probability_calibration


@pytest.fixture()
def waltons():
    return load_waltons()[["T", "E"]].iloc[:50]


@pytest.mark.skipif("DISPLAY" not in os.environ, reason="requires display")
class TestPlotting:
    @pytest.fixture
    def kmf(self):
        return KaplanMeierFitter()

    def setup_method(self, method):
        pytest.importorskip("matplotlib")
        from matplotlib import pyplot as plt

        self.plt = plt

    def test_parametric_univariate_fitters_has_hazard_plotting_methods(self, block, known_parametric_univariate_fitters):
        positive_sample_lifetimes = np.arange(1, 100)
        for fitter in known_parametric_univariate_fitters:
            f = fitter().fit(positive_sample_lifetimes)
            assert f.plot_hazard() is not None
        self.plt.title("test_parametric_univariate_fitters_has_hazard_plotting_methods")
        self.plt.show(block=block)

    def test_parametric_univaraite_fitters_has_cumhazard_plotting_methods(self, block, known_parametric_univariate_fitters):
        positive_sample_lifetimes = np.arange(1, 100)
        for fitter in known_parametric_univariate_fitters:
            f = fitter().fit(positive_sample_lifetimes)
            assert f.plot_cumulative_hazard() is not None

        self.plt.title("test_parametric_univaraite_fitters_has_cumhazard_plotting_methods")
        self.plt.show(block=block)

    def test_parametric_univariate_fitters_has_survival_plotting_methods(self, block, known_parametric_univariate_fitters):
        positive_sample_lifetimes = np.arange(1, 100)
        for fitter in known_parametric_univariate_fitters:
            f = fitter().fit(positive_sample_lifetimes)
            assert f.plot_survival_function() is not None

        self.plt.title("test_parametric_univariate_fitters_has_survival_plotting_methods")
        self.plt.show(block=block)

    def test_negative_times_still_plots(self, block, kmf):
        n = 40
        T = np.linspace(-2, 3, n)
        C = np.random.randint(2, size=n)
        kmf.fit(T, C)
        ax = kmf.plot()
        self.plt.title("test_negative_times_still_plots")
        self.plt.show(block=block)
        return

    def test_kmf_plotting(self, block, kmf):
        data1 = np.random.exponential(10, size=(100))
        data2 = np.random.exponential(2, size=(200, 1))
        data3 = np.random.exponential(4, size=(500, 1))
        kmf.fit(data1, label="test label 1")
        ax = kmf.plot()
        kmf.fit(data2, label="test label 2")
        kmf.plot(ax=ax)
        kmf.fit(data3, label="test label 3")
        kmf.plot(ax=ax)
        self.plt.title("test_kmf_plotting")
        self.plt.show(block=block)
        return

    def test_kmf_with_risk_counts(self, block, kmf):
        data1 = np.random.exponential(10, size=(100))
        kmf.fit(data1)
        kmf.plot(at_risk_counts=True)
        self.plt.title("test_kmf_with_risk_counts")
        self.plt.show(block=block)

    def test_kmf_add_at_risk_counts_with_subplot(self, block, kmf):
        T = np.random.exponential(10, size=(100))
        E = np.random.binomial(1, 0.8, size=(100))
        kmf.fit(T, E)

        fig = self.plt.figure()
        axes = fig.subplots(1, 2)
        kmf.plot(ax=axes[0])
        add_at_risk_counts(kmf, ax=axes[0])
        kmf.plot(ax=axes[1])

        self.plt.title("test_kmf_add_at_risk_counts_with_subplot")
        self.plt.show(block=block)

    def test_kmf_add_at_risk_counts_with_specific_rows(self, block, kmf):
        T = np.random.exponential(10, size=(100))
        E = np.random.binomial(1, 0.8, size=(100))
        kmf.fit(T, E)

        fig = self.plt.figure()
        ax = fig.subplots(1, 1)
        kmf.plot(ax=ax)
        add_at_risk_counts(kmf, ax=ax, rows_to_show=["Censored", "At risk"])

        self.plt.title("test_kmf_add_at_risk_counts_with_specific_rows")
        self.plt.show(block=block)

    def test_kmf_add_at_risk_counts_with_custom_subplot(self, block, kmf):
        # https://github.com/CamDavidsonPilon/lifelines/issues/991#issuecomment-614427882
        import lifelines
        import matplotlib as mpl
        from lifelines.datasets import load_waltons

        plt = self.plt
        waltons = load_waltons()
        ix = waltons["group"] == "control"

        img_no = 3

        height = 4 * img_no
        half_inch = 0.5 / height  # in percent height
        _fig = plt.figure(figsize=(6, height), dpi=100)
        gs = mpl.gridspec.GridSpec(img_no, 1)
        # plt.subplots_adjust(left=0.08, right=0.98, bottom=half_inch, top=1 - half_inch)

        for i in range(img_no):
            ax = plt.subplot(gs[i, 0])
            kmf_control = lifelines.KaplanMeierFitter()
            ax = kmf_control.fit(waltons.loc[ix]["T"], waltons.loc[ix]["E"], label="control").plot(ax=ax)
            kmf_exp = lifelines.KaplanMeierFitter()
            ax = kmf_exp.fit(waltons.loc[~ix]["T"], waltons.loc[~ix]["E"], label="exp").plot(ax=ax)
            ax = lifelines.plotting.add_at_risk_counts(kmf_exp, kmf_control, ax=ax)

        plt.subplots_adjust(hspace=0.6)
        plt.title("test_kmf_add_at_risk_counts_with_custom_subplot")
        plt.show(block=block)

    def test_naf_plotting_with_custom_colours(self, block):
        data1 = np.random.exponential(5, size=(200, 1))
        data2 = np.random.exponential(1, size=(500))
        naf = NelsonAalenFitter()
        naf.fit(data1)
        ax = naf.plot(color="r")
        naf.fit(data2)
        naf.plot(ax=ax, color="k")
        self.plt.title("test_naf_plotting_with_custom_coloirs")
        self.plt.show(block=block)
        return

    def test_ajf_plotting(self, block):
        E = [0, 1, 1, 2, 2, 0]
        T = [1, 2, 3, 4, 5, 6]
        ajf = AalenJohansenFitter().fit(T, E, event_of_interest=1)
        ajf.plot()
        self.plt.title("test_ajf_plotting")
        self.plt.show(block=block)
        return

    def test_ajf_plotting_no_confidence_intervals(self, block):
        E = [0, 1, 1, 2, 2, 0]
        T = [1, 2, 3, 4, 5, 6]
        ajf = AalenJohansenFitter(calculate_variance=False).fit(T, E, event_of_interest=1)
        ajf.plot(ci_show=False)
        self.plt.title("test_ajf_plotting_no_confidence_intervals")
        self.plt.show(block=block)
        return

    def test_ajf_plotting_with_add_count_at_risk(self, block):
        E = [0, 1, 1, 2, 2, 0]
        T = [1, 2, 3, 4, 5, 6]
        ajf = AalenJohansenFitter().fit(T, E, event_of_interest=1)
        ajf.plot(at_risk_counts=True)
        self.plt.title("test_ajf_plotting_with_add_count_at_risk")
        self.plt.show(block=block)
        return

    def test_aalen_additive_plot(self, block):
        # this is a visual test of the fitting the cumulative
        # hazards.
        n = 2500
        d = 3
        timeline = np.linspace(0, 70, 10000)
        hz, coef, X = generate_hazard_rates(n, d, timeline)
        T = generate_random_lifetimes(hz, timeline)
        T[np.isinf(T)] = 10
        C = np.random.binomial(1, 1.0, size=n)
        X["T"] = T
        X["E"] = C

        # fit the aaf, no intercept as it is already built into X, X[2] is ones
        aaf = AalenAdditiveFitter(coef_penalizer=0.1, fit_intercept=False)

        aaf.fit(X, "T", "E")
        ax = aaf.plot(iloc=slice(0, aaf.cumulative_hazards_.shape[0] - 100))
        ax.set_xlabel("time")
        ax.set_title("test_aalen_additive_plot")
        self.plt.show(block=block)
        return

    def test_kmf_with_interval_censoring_plotting(self, block):
        kmf = KaplanMeierFitter()
        left, right = load_diabetes()["left"], load_diabetes()["right"]
        kmf.fit_interval_censoring(left, right)
        kmf.plot(color="r")
        self.plt.show(block=block)
        return

    def test_aalen_additive_smoothed_plot(self, block):
        # this is a visual test of the fitting the cumulative
        # hazards.
        n = 2500
        d = 3
        timeline = np.linspace(0, 150, 5000)
        hz, coef, X = generate_hazard_rates(n, d, timeline)
        T = generate_random_lifetimes(hz, timeline) + 0.1 * np.random.uniform(size=(n, 1))
        C = np.random.binomial(1, 0.8, size=n)
        X["T"] = T
        X["E"] = C

        # fit the aaf, no intercept as it is already built into X, X[2] is ones
        aaf = AalenAdditiveFitter(coef_penalizer=0.1, fit_intercept=False)
        aaf.fit(X, "T", "E")
        ax = aaf.smoothed_hazards_(1).iloc[0 : aaf.cumulative_hazards_.shape[0] - 500].plot()
        ax.set_xlabel("time")
        ax.set_title("test_aalen_additive_smoothed_plot")
        self.plt.show(block=block)
        return

    def test_naf_plotting_slice(self, block):
        data1 = np.random.exponential(5, size=(200, 1))
        data2 = np.random.exponential(1, size=(200, 1))
        naf = NelsonAalenFitter()
        naf.fit(data1)
        ax = naf.plot(loc=slice(0, None))
        naf.fit(data2)
        naf.plot(ax=ax, ci_force_lines=True, iloc=slice(100, 180))
        self.plt.title("test_naf_plotting_slice")
        self.plt.show(block=block)
        return

    def test_plot_lifetimes_calendar(self, block, waltons):
        T, E = waltons["T"], waltons["E"]
        current = 10
        birthtimes = current * np.random.uniform(size=(T.shape[0],))
        ax = plot_lifetimes(T, event_observed=E, entry=birthtimes)
        assert ax is not None
        self.plt.title("test_plot_lifetimes_calendar")
        self.plt.show(block=block)

    def test_plot_lifetimes_left_truncation(self, block, waltons):
        T, E = waltons["T"], waltons["E"]
        N = 20
        current = 10

        birthtimes = current * np.random.uniform(size=(T.shape[0],))
        ax = plot_lifetimes(T, event_observed=E, entry=birthtimes, left_truncated=True)
        assert ax is not None
        self.plt.title("test_plot_lifetimes_left_truncation")
        self.plt.show(block=block)

    def test_MACS_data_with_plot_lifetimes(self, block):
        df = load_multicenter_aids_cohort_study()

        plot_lifetimes(
            df["T"] - df["W"],
            event_observed=df["D"],
            entry=df["W"],
            event_observed_color="#383838",
            event_censored_color="#383838",
            left_truncated=True,
        )
        self.plt.ylabel("Patient Number")
        self.plt.xlabel("Years from AIDS diagnosis")
        self.plt.title("test_MACS_data_with_plot_lifetimes")
        self.plt.show(block=block)

    def test_plot_lifetimes_relative(self, block, waltons):
        T, E = waltons["T"], waltons["E"]
        ax = plot_lifetimes(T, event_observed=E)
        assert ax is not None
        self.plt.title("test_plot_lifetimes_relative")
        self.plt.show(block=block)

    def test_naf_plot_cumulative_hazard(self, block):
        data1 = np.random.exponential(5, size=(200, 1))
        naf = NelsonAalenFitter()
        naf.fit(data1)
        ax = naf.plot()
        naf.plot_cumulative_hazard(ax=ax, ci_force_lines=True)
        self.plt.title("I should have plotted the same thing, but different styles + color!")
        self.plt.show(block=block)
        return

    def test_naf_plot_cumulative_hazard_bandwidth_2(self, block):
        data1 = np.random.exponential(5, size=(2000, 1))
        naf = NelsonAalenFitter()
        naf.fit(data1)
        naf.plot_hazard(bandwidth=1.0, loc=slice(0, 7.0))
        self.plt.title("test_naf_plot_cumulative_hazard_bandwidth_2")
        self.plt.show(block=block)
        return

    def test_naf_plot_cumulative_hazard_bandwith_1(self, block):
        data1 = np.random.exponential(5, size=(2000, 1)) ** 2
        naf = NelsonAalenFitter()
        naf.fit(data1)
        naf.plot_hazard(bandwidth=5.0, iloc=slice(0, 1700))
        self.plt.title("test_naf_plot_cumulative_hazard_bandwith_1")
        self.plt.show(block=block)
        return

    def test_breslow_fleming_harrington_plotting(self, block):
        T = 50 * np.random.exponential(1, size=(200, 1)) ** 2
        bf = BreslowFlemingHarringtonFitter().fit(T)
        bf.plot()
        self.plt.title("test_breslow_fleming_harrington_plotting")
        self.plt.show(block=block)
        return

    def test_weibull_plotting(self, block):
        T = 50 * np.random.exponential(1, size=(200, 1)) ** 2
        wf = WeibullFitter().fit(T, timeline=np.linspace(0, 5, 100))
        wf.plot_hazard()
        self.plt.title("test_weibull_plotting:hazard")
        self.plt.show(block=block)

        wf.plot_cumulative_hazard()
        self.plt.title("test_weibull_plotting:cumulative_hazard")
        self.plt.show(block=block)
        return

    def test_parametric_plotting_with_show_censors(self, block):
        n = 200
        T = (np.sqrt(50) * np.random.exponential(1, size=n)) ** 2
        E = T < 100
        T = np.minimum(T, 100)

        wf = WeibullFitter().fit(T, E)
        wf.plot_density(show_censors=True)
        wf.plot_cumulative_density(show_censors=True)

        self.plt.title("test_parametric_plotting_with_show_censors:cumulative_density")
        self.plt.show(block=block)

        wf.plot_survival_function(show_censors=True)
        self.plt.title("test_parametric_plotting_with_show_censors:survival_function")
        self.plt.show(block=block)

        wf.plot_cumulative_hazard(show_censors=True)
        self.plt.title("test_parametric_plotting_with_show_censors:cumulative_hazard")
        self.plt.show(block=block)

        wf.plot_density(show_censors=True)
        self.plt.title("test_parametric_plotting_with_show_censors:density")
        self.plt.show(block=block)
        return

    def test_label_can_be_changed_on_univariate_fitters(self, block):
        T = np.random.exponential(5, size=(2000, 1)) ** 2
        wf = WeibullFitter().fit(T, timeline=np.linspace(0, 5))
        ax = wf.plot_hazard(label="abc")

        wf.plot_cumulative_hazard(ax=ax, label="123")
        self.plt.title("test_label_can_be_changed_on_univariate_fitters")
        self.plt.show(block=block)
        return

    def test_show_censor_with_discrete_date(self, block, kmf):
        T = np.random.binomial(20, 0.1, size=100)
        C = np.random.binomial(1, 0.8, size=100)
        kmf.fit(T, C).plot(show_censors=True)
        self.plt.title("test_show_censor_with_discrete_date")
        self.plt.show(block=block)
        return

    def test_show_censor_with_index_0(self, block, kmf):
        T = np.random.binomial(20, 0.9, size=100)  # lifelines should auto put a 0 in.
        C = np.random.binomial(1, 0.8, size=100)
        kmf.fit(T, C).plot(show_censors=True)
        self.plt.title("test_show_censor_with_index_0")
        self.plt.show(block=block)
        return

    def test_flat_style_with_custom_censor_styles(self, block, kmf):
        data1 = np.random.exponential(10, size=200)
        E = np.random.rand(200) < 0.8
        kmf.fit(data1, E, label="test label 1")
        kmf.plot(ci_force_lines=True, show_censors=True, censor_styles={"marker": "|", "mew": 1, "ms": 10})
        self.plt.title("test_flat_style_no_censor")
        self.plt.show(block=block)
        return

    def test_loglogs_plot(self, block, kmf):
        data1 = np.random.exponential(10, size=200)
        data2 = np.random.exponential(5, size=200)
        kmf.fit(data1, label="test label 1")
        ax = kmf.plot_loglogs()

        kmf.fit(data2, label="test label 2")
        ax = kmf.plot_loglogs(ax=ax)

        self.plt.title("test_loglogs_plot")
        self.plt.show(block=block)
        return

    def test_seaborn_doesnt_cause_kmf_plot_error(self, block, kmf, capsys):
        import seaborn as sns

        df = load_waltons()

        T = df["T"]
        E = df["E"]

        kmf = KaplanMeierFitter()
        kmf.fit(T, event_observed=E)
        kmf.plot()

        self.plt.title("test_seaborn_doesnt_cause_kmf_plot_error")
        self.plt.show(block=block)
        _, err = capsys.readouterr()
        assert err == ""

    def test_coxph_plotting(self, block):
        df = load_regression_dataset()
        cp = CoxPHFitter()
        cp.fit(df, "T", "E")
        cp.plot()
        self.plt.title("test_coxph_plotting")
        self.plt.show(block=block)

    def test_coxph_plotting_with_hazards_ratios(self, block):
        df = load_regression_dataset()
        cp = CoxPHFitter()
        cp.fit(df, "T", "E")
        cp.plot(hazard_ratios=True)
        self.plt.title("test_coxph_plotting")
        self.plt.show(block=block)

    def test_coxph_plotting_with_subset_of_columns(self, block):
        df = load_regression_dataset()
        cp = CoxPHFitter()
        cp.fit(df, "T", "E")
        cp.plot(columns=["var1", "var2"])
        self.plt.title("test_coxph_plotting_with_subset_of_columns")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome(self, block):
        df = load_rossi()
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest")
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_cumulative_hazard(self, block):
        df = load_rossi()
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest")
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80], y="cumulative_hazard")
        self.plt.title("test_coxph_plot_partial_effects_on_outcome")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_strata(self, block):
        df = load_rossi()
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest", strata=["wexp"])
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_strata")
        self.plt.show(block=block)

    def test_aft_plot_partial_effects_on_outcome_with_categorical(self, block):
        df = load_rossi()
        df["cat"] = np.random.choice(["a", "b", "c"], size=df.shape[0])
        aft = WeibullAFTFitter()
        aft.fit(df, "week", "arrest", formula="cat + age + fin")
        aft.plot_partial_effects_on_outcome("cat", values=["a", "b", "c"])
        self.plt.title("test_aft_plot_partial_effects_on_outcome_with_categorical")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_strata_and_complicated_dtypes(self, block):
        # from https://github.com/CamDavidsonPilon/lifelines/blob/master/examples/Customer%20Churn.ipynb
        churn_data = pd.read_csv(
            "https://raw.githubusercontent.com/"
            "treselle-systems/customer_churn_analysis/"
            "master/WA_Fn-UseC_-Telco-Customer-Churn.csv"
        )
        churn_data = churn_data.set_index("customerID")
        churn_data = churn_data.drop(["TotalCharges"], axis=1)

        churn_data = churn_data.applymap(lambda x: "No" if str(x).startswith("No ") else x)
        churn_data["Churn"] = churn_data["Churn"] == "Yes"
        strata_cols = ["InternetService"]

        cph = CoxPHFitter().fit(
            churn_data,
            "tenure",
            "Churn",
            formula="gender + SeniorCitizen + Partner + Dependents  + MultipleLines + OnlineSecurity + OnlineBackup + DeviceProtection + TechSupport + Contract + PaperlessBilling + PaymentMethod + MonthlyCharges",
            strata=strata_cols,
        )
        cph.plot_partial_effects_on_outcome("Contract", values=["Month-to-month", "One year", "Two year"], plot_baseline=False)
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_strata_and_complicated_dtypes")
        self.plt.show(block=block)

    def test_spline_coxph_plot_partial_effects_on_outcome_with_strata(self, block):
        df = load_rossi()
        cp = CoxPHFitter(baseline_estimation_method="spline", n_baseline_knots=2)
        cp.fit(df, "week", "arrest", strata=["wexp"])
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.title("test_spline_coxph_plot_partial_effects_on_outcome_with_strata")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_single_strata(self, block):
        df = load_rossi()
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest", strata="paro")
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_strata")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_nonnumeric_strata(self, block):
        df = load_rossi()
        df["strata"] = np.random.choice(["A", "B"], size=df.shape[0])
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest", strata="strata")
        cp.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_single_strata")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_multiple_variables(self, block):
        df = load_rossi()
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest")
        cp.plot_partial_effects_on_outcome(["age", "prio"], [[10, 0], [50, 10], [80, 90]])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_multiple_variables")
        self.plt.show(block=block)

    def test_coxph_plot_partial_effects_on_outcome_with_multiple_variables_and_strata(self, block):
        df = load_rossi()
        df["strata"] = np.random.choice(["A", "B"], size=df.shape[0])
        cp = CoxPHFitter()
        cp.fit(df, "week", "arrest", strata="strata")
        cp.plot_partial_effects_on_outcome(["age", "prio"], [[10, 0], [50, 10], [80, 90]])
        self.plt.title("test_coxph_plot_partial_effects_on_outcome_with_multiple_variables_and_strata")
        self.plt.show(block=block)

    def test_coxtv_plotting_with_subset_of_columns(self, block):
        df = load_stanford_heart_transplants()
        ctv = CoxTimeVaryingFitter()
        ctv.fit(df, id_col="id", event_col="event")
        ctv.plot(columns=["age", "year"])
        self.plt.title("test_coxtv_plotting_with_subset_of_columns")
        self.plt.show(block=block)

    def test_coxtv_plotting(self, block):
        df = load_stanford_heart_transplants()
        ctv = CoxTimeVaryingFitter()
        ctv.fit(df, id_col="id", event_col="event")
        ctv.plot(fmt="o")
        self.plt.title("test_coxtv_plotting")
        self.plt.show(block=block)

    def test_kmf_left_censorship_plots(self, block):
        kmf = KaplanMeierFitter()
        lcd_dataset = load_lcd()
        alluvial_fan = lcd_dataset.loc[lcd_dataset["group"] == "alluvial_fan"]
        basin_trough = lcd_dataset.loc[lcd_dataset["group"] == "basin_trough"]
        kmf.fit_left_censoring(alluvial_fan["T"], alluvial_fan["E"], label="alluvial_fan")
        ax = kmf.plot()

        kmf.fit_left_censoring(basin_trough["T"], basin_trough["E"], label="basin_trough")
        ax = kmf.plot(ax=ax)
        self.plt.title("test_kmf_left_censorship_plots")
        self.plt.show(block=block)
        return

    def test_aalen_additive_fit_no_censor(self, block):
        n = 2500
        d = 6
        timeline = np.linspace(0, 70, 10000)
        hz, coef, X = generate_hazard_rates(n, d, timeline)
        X.columns = coef.columns
        cumulative_hazards = pd.DataFrame(cumulative_integral(coef.values, timeline), index=timeline, columns=coef.columns)
        T = generate_random_lifetimes(hz, timeline)
        X["T"] = T
        X["E"] = np.random.binomial(1, 1, n)
        X[np.isinf(X)] = 10
        aaf = AalenAdditiveFitter()
        aaf.fit(X, "T", "E")

        for i in range(d + 1):
            ax = self.plt.subplot(d + 1, 1, i + 1)
            col = cumulative_hazards.columns[i]
            ax = cumulative_hazards[col].loc[:15].plot(ax=ax)
            ax = aaf.plot(loc=slice(0, 15), ax=ax, columns=[col])
        self.plt.title("test_aalen_additive_fit_no_censor")
        self.plt.show(block=block)
        return

    def test_aalen_additive_fit_with_censor(self, block):
        n = 2500
        d = 6
        timeline = np.linspace(0, 70, 10000)
        hz, coef, X = generate_hazard_rates(n, d, timeline)
        X.columns = coef.columns
        cumulative_hazards = pd.DataFrame(cumulative_integral(coef.values, timeline), index=timeline, columns=coef.columns)
        T = generate_random_lifetimes(hz, timeline)
        T[np.isinf(T)] = 10
        X["T"] = T
        X["E"] = np.random.binomial(1, 0.99, n)

        aaf = AalenAdditiveFitter()
        aaf.fit(X, "T", "E")

        for i in range(d + 1):
            ax = self.plt.subplot(d + 1, 1, i + 1)
            col = cumulative_hazards.columns[i]
            ax = cumulative_hazards[col].loc[:15].plot(ax=ax)
            ax = aaf.plot(loc=slice(0, 15), ax=ax, columns=[col])
        self.plt.title("test_aalen_additive_fit_with_censor")
        self.plt.show(block=block)
        return

    def test_weibull_aft_plotting(self, block):
        df = load_regression_dataset()
        aft = WeibullAFTFitter()
        aft.fit(df, "T", "E")
        aft.plot()
        self.plt.tight_layout()
        self.plt.title("test_weibull_aft_plotting")
        self.plt.show(block=block)

    def test_weibull_aft_plotting_with_subset_of_columns(self, block):
        df = load_regression_dataset()
        aft = WeibullAFTFitter()
        aft.fit(df, "T", "E")
        aft.plot(columns=["var1", "var2"])
        self.plt.tight_layout()
        self.plt.title("test_weibull_aft_plotting_with_subset_of_columns")
        self.plt.show(block=block)

    def test_weibull_aft_plot_partial_effects_on_outcome(self, block):
        df = load_rossi()
        aft = WeibullAFTFitter()
        aft.fit(df, "week", "arrest")
        aft.plot_partial_effects_on_outcome("age", [10, 50, 80])
        self.plt.tight_layout()
        self.plt.title("test_weibull_aft_plot_partial_effects_on_outcome")
        self.plt.show(block=block)

    def test_weibull_aft_plot_partial_effects_on_outcome_with_multiple_columns(self, block):
        df = load_rossi()
        aft = WeibullAFTFitter()
        aft.fit(df, "week", "arrest")
        aft.plot_partial_effects_on_outcome(["age", "prio"], [[10, 0], [50, 10], [80, 50]])
        self.plt.tight_layout()
        self.plt.title("test_weibull_aft_plot_partial_effects_on_outcome_with_multiple_columns")
        self.plt.show(block=block)

    def test_left_censorship_cdf_plots(self, block):
        df = load_nh4()
        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit_left_censoring(df["NH4.mg.per.L"], ~df["Censored"])
            ax = cdf_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_left_censorship_cdf_plots")
        self.plt.show(block=block)

    def test_right_censorship_cdf_plots(self, block):
        df = load_rossi()
        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit(df["week"], df["arrest"])
            ax = cdf_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_right_censorship_cdf_plots")
        self.plt.show(block=block)

    def test_qq_plot_left_censoring(self, block):
        df = load_nh4()
        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit_left_censoring(df["NH4.mg.per.L"], ~df["Censored"])
            ax = qq_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_qq_plot_left_censoring")
        self.plt.show(block=block)

    def test_qq_plot_left_censoring2(self, block):
        df = load_lcd()
        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit_left_censoring(df["T"], df["E"])
            ax = qq_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_qq_plot_left_censoring2")
        self.plt.show(block=block)

    def test_qq_plot_left_censoring_with_known_distribution(self, block):
        N = 300
        T_actual = scipy.stats.fisk(8, 0, 1).rvs(N)

        MIN_0 = np.percentile(T_actual, 5)
        MIN_1 = np.percentile(T_actual, 10)

        T = T_actual.copy()
        ix = np.random.randint(3, size=N)

        T = np.where(ix == 0, np.maximum(T, MIN_0), T)
        T = np.where(ix == 1, np.maximum(T, MIN_1), T)
        E = T_actual == T

        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit_left_censoring(T, E)
            ax = qq_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_qq_plot_left_censoring_with_known_distribution")
        self.plt.show(block=block)

    def test_qq_plot_with_weights_and_entry(self, block):
        from lifelines.utils import survival_events_from_table

        df = pd.DataFrame(index=[60, 171, 263, 427, 505, 639])
        df["death"] = [1, 1, 1, 0, 1, 0]
        df["censored"] = [0, 0, 0, 3, 0, 330]
        T, E, W = survival_events_from_table(df, observed_deaths_col="death", censored_col="censored")
        wf = WeibullFitter().fit(T, E, weights=W, entry=0.0001 * np.ones_like(T))
        ax = qq_plot(wf)
        self.plt.suptitle("test_qq_plot_with_weights_and_entry")
        self.plt.show(block=block)

    def test_qq_plot_right_censoring_with_known_distribution(self, block):
        N = 3000
        T_actual = scipy.stats.fisk(8, 0, 1).rvs(N)
        C = scipy.stats.fisk(8, 0, 1).rvs(N)
        E = T_actual < C
        T = np.minimum(T_actual, C)

        fig, axes = self.plt.subplots(2, 2, figsize=(9, 5))
        axes = axes.reshape(4)
        for i, model in enumerate([WeibullFitter(), LogNormalFitter(), LogLogisticFitter(), ExponentialFitter()]):
            model.fit(T, E)
            ax = qq_plot(model, ax=axes[i])
            assert ax is not None
        self.plt.suptitle("test_qq_plot_right_censoring_with_known_distribution")
        self.plt.show(block=block)

    def test_rmst_plot_with_single_model(self, block):
        waltons = load_waltons()
        kmf = KaplanMeierFitter().fit(waltons["T"], waltons["E"])

        rmst_plot(kmf, t=40.0)
        self.plt.title("test_rmst_plot_with_single_model")
        self.plt.show(block=block)

    def test_rmst_plot_with_two_model(self, block):
        waltons = load_waltons()
        ix = waltons["group"] == "control"
        kmf_con = KaplanMeierFitter().fit(waltons.loc[ix]["T"], waltons.loc[ix]["E"], label="control")
        kmf_exp = KaplanMeierFitter().fit(waltons.loc[~ix]["T"], waltons.loc[~ix]["E"], label="exp")

        rmst_plot(kmf_con, model2=kmf_exp, t=40.0)
        self.plt.title("test_rmst_plot_with_two_model")
        self.plt.show(block=block)

    def test_hide_ci_from_legend(self, block):
        waltons = load_waltons()
        kmf = KaplanMeierFitter().fit(waltons["T"], waltons["E"])
        ax = kmf.plot(ci_show=True, ci_only_lines=True, ci_legend=False)
        ax.legend(title="Legend title")
        self.plt.title("test_hide_ci_from_legend")
        self.plt.show(block=block)

    def test_logx_plotting(self, block):
        waltons = load_waltons()
        kmf = KaplanMeierFitter().fit(np.exp(waltons["T"]), waltons["E"], timeline=np.logspace(0, 40))
        ax = kmf.plot(logx=True)

        wf = WeibullFitter().fit(np.exp(waltons["T"]), waltons["E"], timeline=np.logspace(0, 40))
        wf.plot_survival_function(logx=True, ax=ax)

        self.plt.title("test_logx_plotting")
        self.plt.show(block=block)

    def test_survival_probability_calibration(self, block):
        rossi = load_rossi()
        cph = CoxPHFitter().fit(rossi, "week", "arrest")
        survival_probability_calibration(cph, rossi, 25)
        self.plt.title("test_survival_probability_calibration")
        self.plt.show(block=block)

    def test_survival_probability_calibration_on_out_of_sample_data(self, block):
        rossi = load_rossi()
        rossi = rossi.sample(frac=1.0)
        cph = CoxPHFitter().fit(rossi.loc[:300], "week", "arrest")
        survival_probability_calibration(cph, rossi.loc[300:], 25)
        self.plt.title("test_survival_probability_calibration_on_out_of_sample_data")
        self.plt.show(block=block)



# -*- coding: utf-8 -*-
import pytest
from lifelines.fitters.npmle import npmle, is_subset, create_turnbull_intervals, interval, reconstruct_survival_function
from numpy import testing as npt
from lifelines.datasets import load_mice
from lifelines import KaplanMeierFitter
import numpy as np
import pandas as pd


def test_is_subset():
    assert is_subset(interval(4, 4), interval(3, 7))
    assert is_subset(interval(4, 4), interval(4, 7))
    assert is_subset(interval(4, 7), interval(4, 7))
    assert is_subset(interval(4, 7), interval(3.9, 7))
    assert is_subset(interval(4, np.inf), interval(3, np.inf))

    assert not is_subset(interval(4, 9), interval(3.9, 7))
    assert not is_subset(interval(2, 9), interval(3.9, 7))
    assert not is_subset(interval(4, 9), interval(5, 7))


def test_create_turnbull_intervals():
    left, right = zip(*([0, 1], [4, 6], [2, 6], [0, 3], [2, 4], [5, 7]))
    assert create_turnbull_intervals(left, right) == [interval(0, 1), interval(2, 3), interval(4, 4), interval(5, 6)]

    left, right = zip(*[(0, 1)])
    assert create_turnbull_intervals(left, right) == [interval(0, 1)]

    left, right = zip(*[(-1, 1)])
    assert create_turnbull_intervals(left, right) == [interval(-1, 1)]

    left, right = zip(*[(-1, 1), (-2, 2)])
    assert create_turnbull_intervals(left, right) == [interval(-1, 1)]


def test_npmle():
    left, right = [1, 8, 8, 7, 7, 17, 37, 46, 46, 45], [7, 8, 10, 16, 14, np.inf, 44, np.inf, np.inf, np.inf]
    npt.assert_allclose(npmle(left, right, verbose=True)[0], np.array([0.16667016, 0.33332984, 0.125, 0.375]), rtol=1e-4)


def test_npmle_with_weights_is_identical_if_uniform_weights():
    left, right = [1, 8, 8, 7, 7, 17, 37, 46, 46, 45], [7, 8, 10, 16, 14, np.inf, 44, np.inf, np.inf, np.inf]
    weights = 2 * np.ones_like(right)
    npt.assert_allclose(npmle(left, right, verbose=True)[0], np.array([0.16667016, 0.33332984, 0.125, 0.375]), rtol=1e-4)


def test_npmle_with_weights():
    sol = np.array([0.2051282, 0.4102564, 0.0961539, 0.2884615])

    left, right = [1, 8, 8, 7, 7, 17, 37, 46, 46, 45], [7, 8, 10, 16, 14, np.inf, 44, np.inf, np.inf, np.inf]
    weights = np.array([2, 2, 2, 1, 1, 1, 1, 1, 1, 1])
    npt.assert_allclose(npmle(left, right, weights=weights)[0], sol, rtol=1e-4)

    left, right = [1, 1, 8, 8, 8, 8, 7, 7, 17, 37, 46, 46, 45], [7, 7, 8, 8, 10, 10, 16, 14, np.inf, 44, np.inf, np.inf, np.inf]
    npt.assert_allclose(npmle(left, right)[0], sol, rtol=1e-4)


def test_sf_doesnt_return_nans():
    left = [6, 7, 8, 7, 5]
    right = [7, 8, 10, 16, 20]
    results = npmle(left, right)
    npt.assert_allclose(results[1], [interval(7, 7), interval(8, 8)])
    npt.assert_allclose(results[0], [0.5, 0.5])
    sf = reconstruct_survival_function(*results, timeline=[6, 7, 8, 16, 20])
    assert not np.isnan(sf.values).any()


def test_mice_and_optimization_flag():
    df = load_mice()
    results = npmle(df["l"], df["u"], verbose=True, optimize=True)
    npt.assert_allclose(results[0][0], 1 - 0.8571429, rtol=1e-4)
    npt.assert_allclose(results[0][-1], 0.166667, rtol=1e-4)


def test_mice_scipy():
    df = load_mice()
    results = npmle(df["l"], df["u"], verbose=True, fit_method="scipy")
    npt.assert_allclose(results[0][0], 1 - 0.8571429, rtol=1e-4)
    npt.assert_allclose(results[0][-1], 0.166667, rtol=1e-4)


def test_max_lower_bound_less_than_min_upper_bound():
    # https://github.com/CamDavidsonPilon/lifelines/issues/1151
    import numpy as np
    import pandas as pd
    from lifelines import KaplanMeierFitter

    # Data
    np.random.seed(1)
    left0 = np.random.normal(loc=60, scale=2, size=20)
    add_time = np.random.normal(loc=100, scale=2, size=10)
    right1 = left0[0:10] + add_time
    right0 = right1.tolist() + [np.inf] * 10

    # KaplanMeier
    model = KaplanMeierFitter()
    model.fit_interval_censoring(lower_bound=left0, upper_bound=right0)



# -*- coding: utf-8 -*-
import autograd.numpy as np
from autograd.numpy import exp, log
from scipy.special import gammainccinv, gammaincinv
from autograd_gamma import gammaincc, gammainc, gammaln, gammainccln, gammaincln
from lifelines.fitters import KnownModelParametricUnivariateFitter
from lifelines.utils import CensoringType
from lifelines.utils.safe_exp import safe_exp
from autograd.scipy.stats import norm


class GeneralizedGammaFitter(KnownModelParametricUnivariateFitter):
    r"""

    This class implements a Generalized Gamma model for univariate data. The model has parameterized
    form:

    The survival function is:

    .. math::

        S(t)=\left\{  \begin{array}{}
           1-\Gamma_{RL}\left( \frac{1}{{{\lambda }^{2}}};\frac{{e}^{\lambda \left( \frac{\log(t)-\mu }{\sigma} \right)}}{\lambda ^{2}} \right)  \textit{ if } \lambda> 0 \\
              \Gamma_{RL}\left( \frac{1}{{{\lambda }^{2}}};\frac{{e}^{\lambda \left( \frac{\log(t)-\mu }{\sigma} \right)}}{\lambda ^{2}} \right)  \textit{ if } \lambda \le 0 \\
        \end{array} \right.\,\!

    where :math:`\Gamma_{RL}` is the regularized lower incomplete Gamma function.

    This model has the Exponential, Weibull, Gamma and Log-Normal as sub-models, and thus can be used as a way to test which
    model to use:

    1. When :math:`\lambda = 1` and :math:`\sigma = 1`, then the data is Exponential.
    2. When :math:`\lambda = 1` then the data is Weibull.
    3. When :math:`\sigma = \lambda` then the data is Gamma.
    4. When :math:`\lambda = 0` then the data is Log-Normal.
    5. When :math:`\lambda = -1` then the data is Inverse-Weibull.
    6. When :math:`\sigma = -\lambda` then the data is Inverse-Gamma.


    After calling the ``.fit`` method, you have access to properties like: ``cumulative_hazard_``, ``survival_function_``,
    A summary of the fit is available with the method ``print_summary()``.


    Important
    -------------
    The parameterization implemented has :math:`\log\sigma`, thus there is a `ln_sigma_` in the output. Exponentiate this parameter
    to recover :math:`\sigma`.


    Important
    -------------
    This model is experimental. It's API may change in the future. Also, it's convergence is not very stable.


    Parameters
    -----------
    alpha: float, optional (default=0.05)
        the level in the confidence intervals.


    Examples
    --------
    .. code:: python

        from lifelines import GeneralizedGammaFitter
        from lifelines.datasets import load_waltons
        waltons = load_waltons()

        ggf = GeneralizedGammaFitter()
        ggf.fit(waltons['T'], waltons['E'])
        ggf.plot()
        ggf.summary

    Attributes
    ----------
    cumulative_hazard_ : DataFrame
        The estimated cumulative hazard (with custom timeline if provided)
    hazard_ : DataFrame
        The estimated hazard (with custom timeline if provided)
    survival_function_ : DataFrame
        The estimated survival function (with custom timeline if provided)
    cumulative_density_ : DataFrame
        The estimated cumulative density function (with custom timeline if provided)
    density_: DataFrame
        The estimated density function (PDF) (with custom timeline if provided)

    variance_matrix_ : DataFrame
        The variance matrix of the coefficients
    median_survival_time_: float
        The median time to event
    lambda_: float
        The fitted parameter in the model
    rho_: float
        The fitted parameter in the model
    alpha_: float
        The fitted parameter in the model
    durations: array
        The durations provided
    event_observed: array
        The event_observed variable provided
    timeline: array
        The time line to use for plotting and indexing
    entry: array or None
        The entry array provided, or None
    """

    _scipy_fit_method = "SLSQP"
    _fitted_parameter_names = ["mu_", "ln_sigma_", "lambda_"]
    _bounds = [(None, None), (None, None), (None, None)]
    _compare_to_values = np.array([0.0, 0.0, 1.0])

    def _create_initial_point(self, Ts, E, *args):
        if CensoringType.is_right_censoring(self):
            log_data = log(Ts[0])
        elif CensoringType.is_left_censoring(self):
            log_data = log(Ts[1])
        elif CensoringType.is_interval_censoring(self):
            # this fails if Ts[1] == Ts[0], so we add a some fudge factors.
            log_data = log(Ts[1] - Ts[0] + 0.1)
        return np.array([log_data.mean(), log(log_data.std() + 0.01), 0.1])

    def _cumulative_hazard(self, params, times):
        mu_, ln_sigma_, lambda_ = params

        sigma_ = safe_exp(ln_sigma_)
        Z = (log(times) - mu_) / sigma_
        ilambda_2 = 1 / lambda_ ** 2
        clipped_exp = np.clip(safe_exp(lambda_ * Z) * ilambda_2, 1e-300, 1e20)

        if lambda_ > 0:
            v = -gammainccln(ilambda_2, clipped_exp)
        elif lambda_ < 0:
            v = -gammaincln(ilambda_2, clipped_exp)
        else:
            v = -norm.logsf(Z)
        return v

    def _log_hazard(self, params, times):
        mu_, ln_sigma_, lambda_ = params
        ilambda_2 = 1 / lambda_ ** 2
        Z = (log(times) - mu_) / safe_exp(ln_sigma_)
        clipped_exp = np.clip(safe_exp(lambda_ * Z) * ilambda_2, 1e-300, 1e20)
        if lambda_ > 0:
            v = (
                log(lambda_)
                - log(times)
                - ln_sigma_
                - gammaln(ilambda_2)
                + -2 * log(lambda_) * ilambda_2
                - clipped_exp
                + Z / lambda_
                - gammainccln(ilambda_2, clipped_exp)
            )
        elif lambda_ < 0:
            v = (
                log(-lambda_)
                - log(times)
                - ln_sigma_
                - gammaln(ilambda_2)
                - 2 * log(-lambda_) * ilambda_2
                - clipped_exp
                + Z / lambda_
                - gammaincln(ilambda_2, clipped_exp)
            )
        else:
            v = norm.logpdf(Z, loc=0, scale=1) - ln_sigma_ - log(times) - norm.logsf(Z)
        return v

    def percentile(self, p):
        lambda_ = self.lambda_
        sigma_ = exp(self.ln_sigma_)

        if lambda_ > 0:
            return exp(sigma_ * log(gammainccinv(1 / lambda_ ** 2, p) * lambda_ ** 2) / lambda_) * exp(self.mu_)
        return exp(sigma_ * log(gammaincinv(1 / lambda_ ** 2, p) * lambda_ ** 2) / lambda_) * exp(self.mu_)



# -*- coding: utf-8 -*-

import pytest
import numpy as np
import pandas as pd

from lifelines import CoxPHFitter
from lifelines.datasets import load_rossi

from lifelines.utils.concordance import concordance_index
from lifelines.utils.concordance import concordance_index as fast_cindex
from lifelines.utils.concordance import naive_concordance_index as slow_cindex


def test_concordance_index_returns_same_after_shifting():
    T = np.array([1, 2, 3, 4, 5, 6])
    T_ = np.array([2, 1, 4, 6, 5, 3])
    assert (
        concordance_index(T, T_)
        == concordance_index(T - 5, T_ - 5)
        == concordance_index(T, T_ - 5)
        == concordance_index(T - 5, T_)
    )


def test_both_concordance_index_function_deal_with_ties_the_same_way():
    actual_times = np.array([1, 1, 2])
    predicted_times = np.array([1, 2, 3])
    obs = np.ones(3)
    assert fast_cindex(actual_times, predicted_times, obs) == slow_cindex(actual_times, predicted_times, obs) == 1.0


def test_both_concordance_index_with_only_censoring_fails_gracefully():
    actual_times = np.array([1, 2, 3])
    predicted_times = np.array([1, 2, 3])
    obs = np.zeros(3)
    with pytest.raises(ZeroDivisionError, match="admissable pairs"):
        fast_cindex(actual_times, predicted_times, obs)

    with pytest.raises(ZeroDivisionError, match="admissable pairs"):
        slow_cindex(actual_times, predicted_times, obs)


def test_concordance_index_function_exits():
    N = 10 * 1000
    actual_times = np.random.exponential(1, size=N)
    predicted_times = np.random.exponential(1, size=N)
    obs = np.ones(N)
    assert fast_cindex(actual_times, predicted_times, obs)


def test_concordance_index_will_not_overflow():
    a = np.arange(65536)
    assert concordance_index(a, a) == 1.0
    b = np.arange(65537)
    assert concordance_index(b, b) == 1.0
    assert concordance_index(b, b[::-1]) == 0.0


def test_concordance_index_fast_is_same_as_slow():
    size = 100
    T = np.random.normal(size=size)
    P = np.random.normal(size=size)
    C = np.random.choice([0, 1], size=size)
    Z = np.zeros_like(T)

    # Hard to imagine these failing
    assert slow_cindex(T, Z, C) == fast_cindex(T, Z, C)
    assert slow_cindex(T, T, C) == fast_cindex(T, T, C)
    # This is the real test though
    assert slow_cindex(T, P, C) == fast_cindex(T, P, C)

    cp = CoxPHFitter()
    df = load_rossi()
    cp.fit(df, duration_col="week", event_col="arrest")

    T = cp.durations.values.ravel()
    P = -cp.predict_partial_hazard(df[df.columns.difference(["week", "arrest"])]).values.ravel()

    E = cp.event_observed.values.ravel()

    assert slow_cindex(T, P, E) == fast_cindex(T, P, E)
